import pprint

class ValidationException(Exception):
    pass

def validate(expected_schema, datum):
    try:
        return validate_ex(expected_schema, datum)
    except ValidationException:
        return False

INT_MIN_VALUE = -(1 << 31)
INT_MAX_VALUE = (1 << 31) - 1
LONG_MIN_VALUE = -(1 << 63)
LONG_MAX_VALUE = (1 << 63) - 1

def validate_ex(expected_schema, datum):
    """Determine if a python datum is an instance of a schema."""
    schema_type = expected_schema.type
    if schema_type == 'null':
        if datum is None:
            return True
        else:
            raise ValidationException("`%s` is not null" % datum)
    elif schema_type == 'boolean':
        if isinstance(datum, bool):
            return True
        else:
            raise ValidationException("`%s` is not boolean" % datum)
    elif schema_type == 'string':
        if isinstance(datum, basestring):
            return True
        else:
            raise ValidationException("`%s` is not string" % datum)
    elif schema_type == 'bytes':
        if isinstance(datum, str):
            return True
        else:
            raise ValidationException("`%s` is not bytes" % datum)
    elif schema_type == 'int':
        if ((isinstance(datum, int) or isinstance(datum, long))
            and INT_MIN_VALUE <= datum <= INT_MAX_VALUE):
            return True
        else:
            raise ValidationException("`%s` is not int" % datum)
    elif schema_type == 'long':
        if ((isinstance(datum, int) or isinstance(datum, long))
            and LONG_MIN_VALUE <= datum <= LONG_MAX_VALUE):
            return True
        else:
            raise ValidationException("`%s` is not long" % datum)
    elif schema_type in ['float', 'double']:
        if (isinstance(datum, int) or isinstance(datum, long)
            or isinstance(datum, float)):
            return True
        else:
            raise ValidationException("`%s` is not float or double" % datum)
    elif schema_type == 'fixed':
        if isinstance(datum, str) and len(datum) == expected_schema.size:
            return True
        else:
            raise ValidationException("`%s` is not fixed" % datum)
    elif schema_type == 'enum':
        if datum in expected_schema.symbols:
            return True
        else:
            raise ValidationException("`%s`\n is not a valid enum symbol, expected\n %s" % (pprint.pformat(datum), pprint.pformat(expected_schema.symbols)))
    elif schema_type == 'array':
        if isinstance(datum, list):
            for i, d in enumerate(datum):
                try:
                    validate_ex(expected_schema.items, d)
                except ValidationException as v:
                    raise ValidationException("%s\n while validating item at position %i `%s`" % (v, i, d))
            return True
        else:
            raise ValidationException("`%s`\n is not a list, expected list of\n %s" % (pprint.pformat(datum), expected_schema.items))
    elif schema_type == 'map':
        if (isinstance(datum, dict) and
            False not in [isinstance(k, basestring) for k in datum.keys()] and
            False not in [validate(expected_schema.values, v) for v in datum.values()]):
            return True
        else:
            raise ValidationException("`%s` is not a valid map value, expected\n %s" % (pprint.pformat(datum), pprint.pformat(expected_schema.values)))
    elif schema_type in ['union', 'error_union']:
        if True in [validate(s, datum) for s in expected_schema.schemas]:
            return True
        else:
            errors = []
            for s in expected_schema.schemas:
                try:
                    validate_ex(s, datum)
                except ValidationException as e:
                    errors.append(str(e))
            raise ValidationException("`%s`\n is not valid, expected one of:\n\n%s\n\n the individual errors are:\n%s" % (pprint.pformat(datum), ",\n\n  ".join([str(s) for s in expected_schema.schemas]), ";\n\n".join(errors)))
    elif schema_type in ['record', 'error', 'request']:
        if not isinstance(datum, dict):
            raise ValidationException("`%s`\n is not a dict" % pprint.pformat(datum))
        try:
            for f in expected_schema.fields:
                validate_ex(f.type, datum.get(f.name))
            return True
        except ValidationException as v:
            raise ValidationException("%s\n while validating field `%s`" % (v, f.name))
    raise ValidationException("Unrecognized schema_type %s" % schema_type)



import os
import json
import copy
import hashlib
import logging
import collections
import requests
import urlparse
import yaml

log = logging.getLogger(__name__)


class NormDict(dict):
    def __init__(self, normalize=unicode):
        super(NormDict, self).__init__()
        self.normalize = normalize

    def __getitem__(self, key):
        return super(NormDict, self).__getitem__(self.normalize(key))

    def __setitem__(self, key, value):
        return super(NormDict, self).__setitem__(self.normalize(key), value)

    def __delitem__(self, key):
        return super(NormDict, self).__delitem__(self.normalize(key))


class Loader(object):
    def __init__(self):
        normalize = lambda url: urlparse.urlsplit(url).geturl()
        self.fetched = NormDict(normalize)
        self.resolved = NormDict(normalize)
        self.resolving = NormDict(normalize)

    def load(self, url, base_url=None):
        base_url = base_url or 'file://%s/' % os.path.abspath('.')
        return self.resolve_ref({'$ref': url}, base_url)

    def resolve_ref(self, obj, base_url):
        ref, mixin, checksum = obj.pop('$ref', None), obj.pop('$mixin', None), obj.pop('$checksum', None)
        ref = ref or mixin
        url = urlparse.urljoin(base_url, ref)
        if url in self.resolved:
            return self.resolved[url]
        if url in self.resolving:
            raise RuntimeError('Circular reference for url %s' % url)
        self.resolving[url] = True
        doc_url, pointer = urlparse.urldefrag(url)
        document = self.fetch(doc_url)
        fragment = copy.deepcopy(resolve_pointer(document, pointer))
        try:
            self.verify_checksum(checksum, fragment)
            if isinstance(fragment, dict) and mixin:
                fragment = dict(obj, **fragment)
            result = self.resolve_all(fragment, doc_url)
        finally:
            del self.resolving[url]
        return result

    def resolve_all(self, document, base_url):
        if isinstance(document, list):
            iterator = enumerate(document)
        elif isinstance(document, dict):
            if '$ref' in document or '$mixin' in document:
                return self.resolve_ref(document, base_url)
            iterator = document.iteritems()
        else:
            return document
        for key, val in iterator:
            document[key] = self.resolve_all(val, base_url)
        return document

    def fetch(self, url):
        if url in self.fetched:
            return self.fetched[url]
        split = urlparse.urlsplit(url)
        scheme, path = split.scheme, split.path

        if scheme in ['http', 'https'] and requests:
            resp = requests.get(url)
            try:
                resp.raise_for_status()
            except Exception as e:
                raise RuntimeError(url, e)
            result = yaml.load(resp.text)
        elif scheme == 'file':
            try:
                with open(path) as fp:
                    result = yaml.load(fp)
            except (OSError, IOError) as e:
                raise RuntimeError('Failed for %s: %s' % (url, e))
        else:
            raise ValueError('Unsupported scheme: %s' % scheme)
        self.fetched[url] = result
        return result

    def verify_checksum(self, checksum, document):
        if not checksum:
            return
        hash_method, hexdigest = checksum.split('$')
        if hexdigest != self.checksum(document, hash_method):
            raise RuntimeError('Checksum does not match: %s' % checksum)

    def checksum(self, document, method='sha1'):
        if method not in ('md5', 'sha1'):
            raise NotImplementedError('Unsupported hash method: %s' % method)
        normalized = json.dumps(document, sort_keys=True, separators=(',', ':'))
        return getattr(hashlib, method)(normalized).hexdigest


POINTER_DEFAULT = object()


def resolve_pointer(document, pointer, default=POINTER_DEFAULT):
    parts = urlparse.unquote(pointer.lstrip('/#')).split('/') \
        if pointer else []
    for part in parts:
        if isinstance(document, collections.Sequence):
            try:
                part = int(part)
            except ValueError:
                pass
        try:
            document = document[part]
        except:
            if default != POINTER_DEFAULT:
                return default
            else:
                raise ValueError('Unresolvable JSON pointer: %r' % pointer)
    return document


loader = Loader()


def to_json(obj, fp=None):
    default = lambda o: (o.__json__() if callable(getattr(o, '__json__', None))
                         else str(o))
    kwargs = dict(default=default, indent=2, sort_keys=True)
    return json.dump(obj, fp, **kwargs) if fp else json.dumps(obj, **kwargs)


def from_url(url, base_url=None):
    return loader.load(url, base_url)



import avro.schema
import os
import json
import validate
import copy

TOOL_CONTEXT_URL = "https://raw.githubusercontent.com/common-workflow-language/common-workflow-language/master/schemas/draft-2/cwl-context.json"
module_dir = os.path.dirname(os.path.abspath(__file__))

class Process(object):
    def __init__(self, toolpath_object, validateAs):
        self.names = avro.schema.Names()
        cwl_avsc = os.path.join(module_dir, 'schemas/draft-2/cwl.avsc')
        with open(cwl_avsc) as f:
            j = json.load(f)
            for t in j:
                avro.schema.make_avsc_object(t, self.names)

        self.tool = toolpath_object
        #if self.tool.get("@context") != TOOL_CONTEXT_URL:
        #    raise Exception("Missing or invalid '@context' field in tool description document, must be %s" % TOOL_CONTEXT_URL)

        # Validate tool documument
        validate.validate_ex(self.names.get_name(validateAs, ""), self.tool)

        # Import schema defs
        self.schemaDefs = {}
        if self.tool.get("schemaDefs"):
            for i in self.tool["schemaDefs"]:
                avro.schema.make_avsc_object(i, self.names)
                self.schemaDefs[i["name"]] = i

        # Build record schema from inputs
        self.inputs_record_schema = {"name": "input_record_schema", "type": "record", "fields": []}
        for i in self.tool["inputs"]:
            c = copy.copy(i)
            c["name"] = c["id"][1:]
            del c["id"]
            self.inputs_record_schema["fields"].append(c)
        avro.schema.make_avsc_object(self.inputs_record_schema, self.names)

        self.outputs_record_schema = {"name": "outputs_record_schema", "type": "record", "fields": []}
        for i in self.tool["outputs"]:
            c = copy.copy(i)
            c["name"] = c["id"][1:]
            del c["id"]
            self.outputs_record_schema["fields"].append(c)
        avro.schema.make_avsc_object(self.outputs_record_schema, self.names)



import job
import draft1tool
import draft2tool
from draft2tool import aslist
from process import Process
import copy
import logging
import random
from ref_resolver import from_url
import os

_logger = logging.getLogger("cwltool")

def makeTool(toolpath_object, basedir):
    if "schema" in toolpath_object:
        return draft1tool.Tool(toolpath_object)
    elif "impl" in toolpath_object and toolpath_object.get("class", "External") == "External":
        return External(toolpath_object, basedir)
    if "class" in toolpath_object:
        if toolpath_object["class"] == "CommandLineTool":
            return draft2tool.CommandLineTool(toolpath_object)
        elif toolpath_object["class"] == "ExpressionTool":
            return draft2tool.ExpressionTool(toolpath_object)
        elif toolpath_object["class"] == "Workflow":
            return Workflow(toolpath_object)
    else:
        raise Exception("Missing 'class' field, expecting one of: Workflow, CommandLineTool, ExpressionTool, External")


def should_fanout(src_type, dest_type):
    if isinstance(src_type, dict):
        if src_type["type"] == "array" and src_type["items"] == dest_type:
            return True
    return False

class WorkflowJob(object):
    def try_make_job(self, s):
        jo = {}
        fanout = None
        for i in s.tool["inputs"]:
            _logger.debug(i)
            if "connect" in i:
                connect = i["connect"]
                if isinstance(connect, list):
                    # Handle multiple inputs
                    if not fanout:
                        fanout = i["id"][1:]
                        jo[i["id"][1:]] = []
                    else:
                        raise Exception("Can only fanout on one port")
                for c in aslist(connect):
                    src = c["source"][1:]
                    if src in self.state:
                        if self.state[src][0]["type"] == i["type"]:
                            if fanout:
                                jo[i["id"][1:]].append(self.state[src][1])
                            else:
                                jo[i["id"][1:]] = self.state[src][1]
                        elif should_fanout(self.state[src][0]["type"], i["type"]):
                            if fanout:
                                if fanout == i["id"][1:]:
                                    jo[i["id"][1:]].extend(self.state[src][1])
                                else:
                                    raise Exception("Can only fanout on one port")
                            else:
                                fanout = i["id"][1:]
                                jo[i["id"][1:]] = self.state[src][1]
                        else:
                            raise Exception("Type mismatch '%s' and '%s'" % (src, i["id"][1:]))
                    else:
                        return None
            elif "default" in i:
                jo[i["id"][1:]] = i["default"]

        _logger.info("Creating job with input: %s", jo)
        if fanout:
            s = Fanout(s, fanout)
        return s.job(jo, self.basedir)

    def run(self, outdir=None, **kwargs):
        for s in self.steps:
            s.completed = False

        run_all = len(self.steps)
        while run_all:
            made_progress = False
            for s in self.steps:
                if not s.completed:
                    job = self.try_make_job(s)
                    if job:
                        (joutdir, output) = job.run(outdir=outdir)
                        for i in s.tool["outputs"]:
                            _logger.info("Job got output: %s", output)
                            if "id" in i:
                                if i["id"][1:] in output:
                                    self.state[i["id"][1:]] = (i, output[i["id"][1:]])
                                else:
                                    raise Exception("Output is missing expected field %s" % i["id"][1:])
                        s.completed = True
                        made_progress = True
                        run_all -= 1
            if not made_progress:
                raise Exception("Deadlocked")

        wo = {}
        for i in self.outputs:
            if "connect" in i:
                src = i["connect"]["source"][1:]
                wo[i["id"][1:]] = self.state[src][1]

        return (outdir, wo)


class Workflow(Process):
    def __init__(self, toolpath_object):
        super(Workflow, self).__init__(toolpath_object, "Workflow")

    def job(self, joborder, basedir, use_container=True):
        wj = WorkflowJob()
        wj.basedir = basedir
        wj.steps = [makeTool(s, basedir) for s in self.tool.get("steps", [])]
        random.shuffle(wj.steps)

        wj.state = {}
        for i in self.tool["inputs"]:
            iid = i["id"][1:]
            if iid in joborder:
                wj.state[iid] = (i, copy.deepcopy(joborder[iid]))
            elif "default" in i:
                wj.state[iid] = (i, copy.deepcopy(i["default"]))
        wj.outputs = self.tool["outputs"]
        return wj

class ExternalJob(object):
    def __init__(self, tool, innerjob):
        self.tool = tool
        self.innerjob = innerjob

    def run(self, **kwargs):
        self.impl = self.tool["impl"]
        (outdir, output) = self.innerjob.run(**kwargs)
        for i in self.tool["outputs"]:
            d = i["def"][len(self.impl)+1:]
            output[i["id"][1:]] = output[d]
            del output[d]

        return (outdir, output)

class External(Process):
    def __init__(self, toolpath_object, basedir):
        self.impl = toolpath_object["impl"]
        self.embedded_tool = makeTool(from_url(os.path.join(basedir, self.impl)), basedir)

        if "id" in toolpath_object:
            self.id = toolpath_object["id"]
        else:
            self.id = "#step_" + str(random.randint(1, 1000000000))

        for i in toolpath_object["inputs"]:
            d = i["def"][len(self.impl):]
            toolid = i.get("id", self.id + "." + d[1:])
            found = False
            for a in self.embedded_tool.tool["inputs"]:
                if a["id"] == d:
                    i.update(a)
                    found = True
            if not found:
                raise Exception("Did not find input '%s' in external process" % (i["def"]))

            i["id"] = toolid

        for i in toolpath_object["outputs"]:
            d = i["def"][len(self.impl):]
            toolid = i["id"]
            found = False
            for a in self.embedded_tool.tool["outputs"]:
                if a["id"] == d:
                    i.update(a)
                    found = True
            if not found:
                raise Exception("Did not find output '%s' in external process" % (i["def"]))

            i["id"] = toolid

        super(External, self).__init__(toolpath_object, "Process")

    def job(self, joborder, basedir, **kwargs):
        for i in self.tool["inputs"]:
            d = i["def"][len(self.impl)+1:]
            joborder[d] = joborder[i["id"][1:]]
            del joborder[i["id"][1:]]

        return ExternalJob(self.tool, self.embedded_tool.job(joborder, basedir, **kwargs))

class FanoutJob(object):
    def __init__(self, outputports, jobs):
        self.outputports = outputports
        self.jobs = jobs

    def run(self, **kwargs):
        outputs = {}
        for outschema in self.outputports:
            outputs[outschema["id"][1:]] = []
        for j in self.jobs:
            (_, out) = j.run(**kwargs)
            for outschema in self.outputports:
                outputs[outschema["id"][1:]].append(out[outschema["id"][1:]])
        return (None, outputs)

class Fanout(object):
    def __init__(self, process, fanout_key):
        self.process = process
        self.fanout_key = fanout_key
        self.outputports = []
        for out in self.process.tool["outputs"]:
            newout = copy.deepcopy(out)
            newout["type"] = {"type": "array", "items": out["type"]}
            self.outputports.append(newout)
        self.tool = {"outputs": self.outputports}

    def job(self, joborder, basedir, **kwargs):
        jobs = []
        for fn in joborder[self.fanout_key]:
            jo = copy.copy(joborder)
            jo[self.fanout_key] = fn
            jobs.append(self.process.job(jo, basedir, **kwargs))
        return FanoutJob(self.outputports, jobs)



import avro.schema
import json
import pprint
import copy
from flatten import flatten
import functools
import os
from pathmapper import PathMapper, DockerPathMapper
import sandboxjs
from job import CommandLineJob
import yaml
import glob
import logging
import hashlib
import random
from process import Process
import validate

_logger = logging.getLogger("cwltool")

CONTENT_LIMIT = 1024 * 1024

module_dir = os.path.dirname(os.path.abspath(__file__))

class Builder(object):
    def jseval(self, expression, context):
        if isinstance(expression, list):
            exp = "{return %s(%s);}" % (expression[0], ",".join([json.dumps(self.do_eval(e)) for e in expression[1:]]))
        elif expression.startswith('{'):
            exp = '{return function()%s();}' % (expression)
        else:
            exp = '{return %s;}' % (expression)
        return sandboxjs.execjs(exp, "var $job = %s; var $self = %s; %s" % (json.dumps(self.job), json.dumps(context), self.jslib))

    def do_eval(self, ex, context=None):
        if isinstance(ex, dict):
            if ex.get("class") == "JavascriptExpression":
                if "value" in ex:
                    return self.jseval(ex["value"], context)
                elif "invoke" in ex:
                    return self.jseval(ex["invoke"], context)
            elif ex.get("id"):
                if ex["id"].startswith("#"):
                    return self.job[ex["id"][1:]]
                else:
                    with open(os.path.join(self.basedir, ex["id"]), "r") as f:
                        return f.read()
        else:
            return ex

    def bind_input(self, schema, datum):
        bindings = []

        # Handle union types
        if isinstance(schema["type"], list):
            success = False
            for t in schema["type"]:
                if t in self.schemaDefs:
                    t = self.schemaDefs[t]
                avsc = avro.schema.make_avsc_object(t, None)
                if validate.validate(avsc, datum):
                    if isinstance(t, basestring):
                        t = {"type": t}
                    bindings.extend(self.bind_input(t, datum))
                    success = True
                    break
            if not success:
                raise ValidationException("'%s' is not a valid union %s" % (datum, schema["type"]))
        elif isinstance(schema["type"], dict):
            bindings.extend(self.bind_input(schema["type"], datum))
        else:
            if schema["type"] in self.schemaDefs:
                schema = self.schemaDefs[schema["type"]]

            if schema["type"] == "record":
                for f in schema["fields"]:
                    if f["name"] in datum:
                        b = self.bind_input(f, datum[f["name"]])
                        for bi in b:
                            bi["position"].append(f["name"])
                        bindings.extend(b)

            if schema["type"] == "map":
                for v in datum:
                    b = self.bind_input(schema["values"], datum[v])
                    for bi in b:
                        bi["position"].insert(0, v)
                    bindings.extend(b)

            if schema["type"] == "array":
                for n, item in enumerate(datum):
                    b = self.bind_input({"type": schema["items"], "commandLineBinding": schema.get("commandLineBinding")}, item)
                    for bi in b:
                        bi["position"].insert(0, n)
                    bindings.extend(b)

            if schema["type"] == "File":
                if schema.get("loadContents"):
                    with open(os.path.join(self.basedir, datum["path"]), "rb") as f:
                        datum["contents"] = f.read(CONTENT_LIMIT)
                self.files.append(datum)

        b = None
        if "commandLineBinding" in schema and isinstance(schema["commandLineBinding"], dict):
            b = copy.copy(schema["commandLineBinding"])

            if b.get("position"):
                b["position"] = [b["position"]]
            else:
                b["position"] = [0]

            # Position to front of the sort key
            for bi in bindings:
                bi["position"] = b["position"] + bi["position"]

            if "valueFrom" in b:
                b["do_eval"] = b["valueFrom"]
            b["valueFrom"] = datum

            if schema["type"] == "File":
                b["is_file"] = True
            bindings.append(b)

        return bindings

    def generate_arg(self, binding):
        value = binding["valueFrom"]
        if "do_eval" in binding:
            value = self.do_eval(binding["do_eval"], value)

        prefix = binding.get("prefix")
        sep = binding.get("separator")

        l = []
        if isinstance(value, list):
            if binding.get("itemSeparator"):
                l = [binding["itemSeparator"].join([str(v) for v in value])]
            elif binding.get("do_eval"):
                return ([prefix] if prefix else []) + value
            elif prefix:
                return [prefix]
            else:
                return []
        elif binding.get("is_file"):
            l = [value["path"]]
        elif isinstance(value, dict):
            return [prefix] if prefix else []
        elif value is True and prefix:
            return [prefix]
        elif value is False or value is None:
            return []
        else:
            l = [value]

        args = []
        for j in l:
            if sep is None or sep == " ":
                args.extend([prefix, str(j)])
            else:
                args.extend([prefix + sep + str(j)])

        return [a for a in args if a is not None]


class Tool(Process):
    def _init_job(self, joborder, basedir):
        # Validate job order
        validate.validate_ex(self.names.get_name("input_record_schema", ""), joborder)

        builder = Builder()
        builder.job = copy.deepcopy(joborder)
        builder.jslib = ''
        builder.basedir = basedir
        builder.files = []
        builder.bindings = []
        builder.schemaDefs = self.schemaDefs

        if self.tool.get("expressionDefs"):
            for ex in self.tool['expressionDefs']:
                builder.jslib += builder.do_eval(ex) + "\n"

        builder.bindings.extend(builder.bind_input(self.inputs_record_schema, builder.job))

        return builder


class ExpressionTool(Tool):
    def __init__(self, toolpath_object):
        super(ExpressionTool, self).__init__(toolpath_object, "ExpressionTool")

    class ExpressionJob(object):
        def run(self, outdir=None, **kwargs):
            return (outdir, self.builder.do_eval(self.script))

    def job(self, joborder, basedir, **kwargs):
        builder = self._init_job(joborder, basedir)

        j = ExpressionTool.ExpressionJob()
        j.builder = builder
        j.script = self.tool["script"]

        return j

def aslist(l):
    if isinstance(l, list):
        return l
    else:
        return [l]

class CommandLineTool(Tool):
    def __init__(self, toolpath_object):
        super(CommandLineTool, self).__init__(toolpath_object, "CommandLineTool")

    def job(self, joborder, basedir, use_container=True):
        builder = self._init_job(joborder, basedir)

        if self.tool["baseCommand"]:
            for n, b in enumerate(aslist(self.tool["baseCommand"])):
                builder.bindings.append({
                    "position": [-1000000, n],
                    "valueFrom": b
                })

        if self.tool.get("arguments"):
            for i, a in enumerate(self.tool["arguments"]):
                if isinstance(a, dict):
                    a = copy.copy(a)
                    if a.get("position"):
                        a["position"] = [a["position"], i]
                    else:
                        a["position"] = [0, i]
                    a["do_eval"] = a["valueFrom"]
                    a["valueFrom"] = None
                    builder.bindings.append(a)
                else:
                    builder.bindings.append({
                        "position": [0, i],
                        "valueFrom": a
                    })

        builder.bindings.sort(key=lambda a: a["position"])

        _logger.debug(pprint.pformat(builder.bindings))
        _logger.debug(pprint.pformat(builder.files))

        reffiles = [f["path"] for f in builder.files]

        j = CommandLineJob()
        j.joborder = builder.job
        j.container = None
        j.stdin = None
        j.stdout = None
        builder.pathmapper = None

        if self.tool.get("stdin"):
            j.stdin = builder.do_eval(self.tool["stdin"])
            if isinstance(j.stdin, dict):
                j.stdin = j.stdin["path"]
            reffiles.append(j.stdin)

        if self.tool.get("stdout"):
            if isinstance(self.tool["stdout"], dict) and "id" in self.tool["stdout"]:
                for out in self.tool.get("outputs", []):
                    if out["id"] == self.tool["stdout"]["id"]:
                        filename = self.tool["stdout"]["id"][1:]
                        j.stdout = filename
                        out["outputBinding"] = out.get("outputBinding", {})
                        out["outputBinding"]["glob"] = filename
                if not j.stdout:
                    raise Exception("stdout refers to invalid output")
            else:
                j.stdout = builder.do_eval(self.tool["stdout"])
            if os.path.isabs(j.stdout):
                raise Exception("stdout must be a relative path")

        j.generatefiles = {}
        for t in self.tool.get("fileDefs", []):
            j.generatefiles[t["filename"]] = builder.do_eval(t["value"])

        j.environment = {}
        for t in self.tool.get("environmentDefs", []):
            j.environment[t["env"]] = builder.do_eval(t["value"])

        reqsAndHints = self.tool.get("requirements", []) + self.tool.get("hints", [])
        for r in reqsAndHints:
            if r["class"] == "DockerRequirement" and use_container:
                j.container = {}
                j.container["type"] = "docker"
                if "dockerPull" in r:
                    j.container["pull"] = r["dockerPull"]
                if "dockerLoad" in r:
                    if r["dockerLoad"].startswith("http"):
                        j.container["load"] = r["dockerLoad"]
                    else:
                        j.container["load"] = os.path.join(basedir, r["dockerLoad"])
                if "dockerImageId" in r:
                    j.container["imageId"] = r["dockerImageId"]
                else:
                    j.container["imageId"] = r["dockerPull"]
                builder.pathmapper = DockerPathMapper(reffiles, basedir)

        if builder.pathmapper is None:
            builder.pathmapper = PathMapper(reffiles, basedir)

        for f in builder.files:
            f["path"] = builder.pathmapper.mapper(f["path"])

        j.command_line = flatten(map(builder.generate_arg, builder.bindings))

        if j.stdin:
            j.stdin = j.stdin if os.path.isabs(j.stdin) else os.path.join(basedir, j.stdin)

        j.pathmapper = builder.pathmapper
        j.collect_outputs = functools.partial(self.collect_output_ports, self.tool["outputs"], builder)

        return j

    def collect_output_ports(self, ports, builder, outdir):
        custom_output = os.path.join(outdir, "output.cwl.json")
        if os.path.exists(custom_output):
            outputdoc = yaml.load(custom_output)
            validate.validate_ex(self.names.get_name("output_record_schema", ""), outputdoc)
            return outputdoc
        ret = {port["id"][1:]: self.collect_output(port, builder, outdir) for port in ports}
        return ret if ret is not None else {}

    def collect_output(self, schema, builder, outdir):
        r = None
        if "outputBinding" in schema:
            binding = schema["outputBinding"]
            if "glob" in binding:
                r = [{"path": g} for g in glob.glob(os.path.join(outdir, binding["glob"]))]
                for files in r:
                    checksum = hashlib.sha1()
                    with open(files["path"], "rb") as f:
                        contents = f.read(CONTENT_LIMIT)
                        if binding.get("loadContents"):
                            files["contents"] = contents
                        filesize = 0
                        while contents != "":
                            checksum.update(contents)
                            filesize += len(contents)
                            contents = f.read(1024*1024)
                    files["checksum"] = "sha1$%s" % checksum.hexdigest()
                    files["size"] = filesize

                if schema["type"] == "array" and schema["items"] == "File":
                    pass
                elif schema["type"] == "File":
                    r = r[0] if r else None
                elif binding.get("loadContents"):
                    r = [v["contents"] for v in r]
                else:
                    r = None

            if "valueFrom" in binding:
                r = builder.do_eval(binding["valueFrom"], r)

        if not r and schema["type"] == "record":
            r = {}
            for f in schema["fields"]:
                r[f["name"]] = self.collect_output(f, builder, outdir)

        return r



import subprocess
import json
import threading

class JavascriptException(Exception):
    pass

def execjs(js, jslib):
    nodejs = subprocess.Popen(["nodejs"], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    fn = "\"use strict\";%s\n(function()%s)()" % (jslib, js if isinstance(js, basestring) and len(js) > 1 and js[0] == '{' else ("{return (%s);}" % js))
    script = "console.log(JSON.stringify(require(\"vm\").runInNewContext(%s, {})))" % json.dumps(fn)

    #print script

    def term():
        try:
            nodejs.terminate()
        except OSError:
            pass

    # Time out after 5 seconds
    tm = threading.Timer(5, term)
    tm.start()

    stdoutdata, stderrdata = nodejs.communicate(script)
    tm.cancel()

    if stderrdata.strip() or nodejs.returncode != 0:
        raise JavascriptException(script + "\n" + stderrdata)
    else:
        return json.loads(stdoutdata)



import os
import pprint
import json
import sandboxjs
import copy
import sys
import jsonschema.exceptions
import random
import requests
import urlparse
import functools
from pathmapper import PathMapper, DockerPathMapper
from job import CommandLineJob
from flatten import flatten

from jsonschema.validators import Draft4Validator
import ref_resolver
from ref_resolver import from_url, resolve_pointer

module_dir = os.path.dirname(os.path.abspath(__file__))

jsonschemapath = os.path.join(module_dir, 'schemas/draft-1/json-schema-draft-04.json')
with open(jsonschemapath) as f:
    jsonschemapath_doc = json.load(f)

ref_resolver.loader.fetched["http://json-schema.org/draft-04/schema"] = jsonschemapath_doc

toolpath = os.path.join(module_dir, 'schemas/draft-1/tool.json')
with open(toolpath) as f:
    tool_schema_doc = json.load(f)
with open(os.path.join(module_dir, 'schemas/draft-1/metaschema.json')) as f:
    metaschema = json.load(f)

SCHEMA_URL_PREFIX = "https://raw.githubusercontent.com/common-workflow-language/common-workflow-language/draft-1/schemas/"
TOOL_SCHEMA_URL = SCHEMA_URL_PREFIX + "tool.json"
METASCHEMA_SCHEMA_URL = SCHEMA_URL_PREFIX + "metaschema.json"

ref_resolver.loader.fetched[TOOL_SCHEMA_URL] = tool_schema_doc
ref_resolver.loader.fetched[METASCHEMA_SCHEMA_URL] = metaschema

tool_schema = Draft4Validator(tool_schema_doc)

def each(l):
    if l is None:
        return []
    if isinstance(l, (list, tuple)):
        return l
    else:
        return [l]

def fix_file_type(t):
    if 'type' in t and t['type'] == "file":
        for a in metaschema["definitions"]["file"]:
            t[a] = metaschema["definitions"]["file"][a]
        t["_type"] = "file"
    for k in t:
        if isinstance(t[k], dict):
            fix_file_type(t[k])

class Builder(object):

    def jseval(self, job=None, expression=None):
        if expression.startswith('{'):
            exp_tpl = '''{
            return function()%s();}
            '''
        else:
            exp_tpl = '''{
            return %s;}
            '''
        exp = exp_tpl % (expression)
        return sandboxjs.execjs(exp, "var $job = %s;%s" % (json.dumps(job), self.jslib))

    def resolve_eval(self, job, v):
        if isinstance(v, dict):
            if "$expr" in v:
                # Support $import of the $expr
                return self.jseval(job, self.resolve_eval(job, v["$expr"]))
            if "$apply" in v:
                # Support $import of the $expr
                ex = ""
                for i, p in enumerate(v["$apply"]):
                    if i == 0:
                        ex += p + "("
                    else:
                        ex += json.dumps(self.resolve_eval(job, p))
                        if i < len(v["$apply"])-1:
                            ex += ","
                ex += ")"
                return self.jseval(job, ex)
            elif "$job" in v:
                return resolve_pointer(job, v["$job"])
            elif "$import" in v:
                # TODO: check checksum
                url = urlparse.urljoin(self.base_url, v["$import"])
                split = urlparse.urlsplit(url)
                scheme, path = split.scheme, split.path
                if scheme in ['http', 'https']:
                    resp = requests.get(url)
                    try:
                        resp.raise_for_status()
                    except Exception as e:
                        raise RuntimeError(url, e)
                    return resp.text
                elif scheme == 'file':
                    try:
                        with open(path) as fp:
                            return fp.read()
                    except (OSError, IOError) as e:
                        raise RuntimeError('Failed for %s: %s' % (url, e))
                else:
                    raise ValueError('Unsupported scheme: %s' % scheme)
        return v

    def adapt_inputs(self, schema, job, inp, key):
        adapters = []

        if 'oneOf' in schema:
            for one in schema["oneOf"]:
                try:
                    Draft4Validator(one).validate(inp)
                    schema = one
                    break
                except jsonschema.exceptions.ValidationError:
                    pass

        if isinstance(inp, dict):
            if "properties" in schema:
                for i in inp:
                    a = self.adapt_inputs(schema["properties"][i], job, inp[i], i)
                    adapters.extend(a)
        elif isinstance(inp, list):
            for n, i in enumerate(inp):
                a = self.adapt_inputs(schema["items"], job, i, format(n, '06'))
                for x in a:
                    x["order"].insert(0, n)
                adapters.extend(a)

        if 'adapter' in schema:
            a = copy.copy(schema['adapter'])

            if "order" in a:
                a["order"] = [a["order"], key]
            else:
                a["order"] = [1000000, key]

            a["schema"] = schema

            for x in adapters:
                x["order"] = a["order"] + x["order"]

            if not 'value' in a and len(adapters) == 0:
                a['value'] = inp

            if len(adapters) == 0 or "value" in a:
                adapters.insert(0, a)

        return adapters

    def to_str(self, schema, value, path_mapper):
        if "$ref" in schema:
            schema = from_url(schema["$ref"], self.ref_base_url)

        if 'oneOf' in schema:
            for a in schema['oneOf']:
                v = self.to_str(a, value, path_mapper)
                if v is not None:
                    return v
            return None
        elif 'type' in schema:
            if schema["type"] == "array" and isinstance(value, list):
                return [self.to_str(schema["items"], v, path_mapper) for v in value]
            elif schema["type"] == "object" and isinstance(value, dict):
                if "path" in value:
                    return path_mapper(value["path"])
                else:
                    raise Exception("Not expecting a dict %s" % (value))
            elif schema["type"] in ("string", "number", "integer"):
                return str(value)
            elif schema["type"] == "boolean":
                # handled specially by adapt()
                return value

        return None

    def find_files(self, adapter, job):
        if "value" in adapter:
            value = self.resolve_eval(job, adapter["value"])
        else:
            return None

        schema = adapter["schema"]

        if "$ref" in schema:
            schema = from_url(schema["$ref"], self.ref_base_url)

        if 'oneOf' in schema:
            for a in schema['oneOf']:
                v = self.find_files(a, value)
                if v is not None:
                    return v
            return None
        elif 'type' in schema:
            if schema["type"] == "array" and isinstance(value, list):
                return [self.find_files({"value": v,
                                    "schema": schema["items"]}, job) for v in value]
            elif schema["type"] == "object" and isinstance(value, dict):
                if "path" in value:
                    return value["path"]
                else:
                    raise Exception("Not expecting a dict %s" % (value))

        return None


    def adapt(self, adapter, job, path_mapper):
        if "value" in adapter:
            value = self.resolve_eval(job, adapter["value"])
        else:
            raise Exception("No value in adapter")

        value = self.to_str(adapter["schema"], value, path_mapper)

        sep = adapter["separator"] if "separator" in adapter else " "

        if 'itemSeparator' in adapter:
            if adapter["prefix"]:
                l = [adapter["prefix"] + adapter['itemSeparator'].join(value)]
            else:
                l = [adapter['itemSeparator'].join(value)]
        elif 'prefix' in adapter:
            l = []
            if value is True:
                l.append(adapter["prefix"])
            elif value is False:
                pass
            else:
                for v in each(value):
                    if sep == " ":
                        l.append(adapter["prefix"])
                        l.append(v)
                    else:
                        l.append(adapter["prefix"] + sep + v)
        else:
            l = [value]

        return l

class Tool(object):
    def __init__(self, toolpath_object):
        self.tool = toolpath_object
        fix_file_type(self.tool)
        if "schema" not in self.tool or self.tool["schema"] != TOOL_SCHEMA_URL:
            raise Exception("Missing or invalid 'schema' field in tool description document, must be %s" % TOOL_SCHEMA_URL)
        tool_schema.validate(self.tool)

    def job(self, joborder, basedir, use_container=True):
        inputs = joborder['inputs']
        Draft4Validator(self.tool['inputs']).validate(inputs)

        adapter = self.tool["adapter"]
        adapters = [{"order": [-1000000],
                     "schema": tool_schema_doc["properties"]["adapter"]["properties"]["baseCmd"],
                     "value": adapter['baseCmd']
                 }]

        builder = Builder()
        builder.base_url = "file:"+os.path.abspath(basedir)+"/"
        builder.ref_base_url = "file:"+toolpath

        requirements = self.tool.get("requirements")
        builder.jslib = ''
        if requirements and 'expressionlib' in requirements:
            for ex in requirements['expressionlib']:
                builder.jslib += builder.resolve_eval(joborder, ex) + "\n"

        if "args" in adapter:
            for i, a in enumerate(adapter["args"]):
                a = copy.copy(a)
                if "order" in a:
                    a["order"] = [a["order"]]
                else:
                    a["order"] = [0]
                a["schema"] = tool_schema_doc["definitions"]["strOrExpr"]
                adapters.append(a)

        adapters.extend(builder.adapt_inputs(self.tool['inputs'], inputs, inputs, ""))

        adapters.sort(key=lambda a: a["order"])

        referenced_files = filter(lambda a: a is not None, flatten(map(lambda a: builder.find_files(a, joborder), adapters)))

        j = CommandLineProcess()
        j.joborder = joborder
        j.container = None

        if 'stdin' in adapter:
            j.stdin = flatten(builder.adapt({"value": adapter['stdin'],
                                              "schema": tool_schema_doc["properties"]["adapter"]["properties"]["stdin"]
                                          }, joborder, None))[0]
            referenced_files.append(j.stdin)
        else:
            j.stdin = None

        if 'stdout' in adapter:
            j.stdout = flatten(builder.adapt({"value": adapter['stdout'],
                                               "schema": tool_schema_doc["properties"]["adapter"]["properties"]["stdout"]
                                           }, joborder, None))[0]

            if os.path.isabs(j.stdout):
                raise Exception("stdout must be a relative path")
        else:
            j.stdout = None

        j.generatefiles = {}
        for t in adapter.get("generatefiles", []):
            j.generatefiles[builder.resolve_eval(inputs, t["name"])] = builder.resolve_eval(inputs, t["value"])

        d = None
        if requirements:
            b = requirements.get("environment")
            if b:
                c = b.get("container")
                if use_container and c:
                    if c.get("type") == "docker":
                        d = DockerPathMapper(referenced_files, basedir)
                        j.container = c

        if d is None:
            d = PathMapper(referenced_files, basedir)

        if j.stdin:
            j.stdin = j.stdin if os.path.isabs(j.stdin) else os.path.join(basedir, j.stdin)

        j.command_line = flatten(map(lambda a: builder.adapt(a, joborder, d.mapper), adapters))

        j.pathmapper = d
        j.collect_outputs = functools.partial(self.collect_outputs, self.tool.get("outputs", {}), joborder)

        return j

    def collect_outputs(self, schema, joborder, outdir):
        result_path = os.path.join(outdir, "result.cwl.json")
        if os.path.isfile(result_path):
            print "Result file found."
            with open(result_path) as fp:
                return yaml.load(fp)

        r = None
        if isinstance(schema, dict):
            if "adapter" in schema:
                adapter = schema["adapter"]
                if "glob" in adapter:
                    r = [{"path": g} for g in glob.glob(os.path.join(outdir, adapter["glob"]))]
                    if not ("type" in schema and schema["type"] == "array"):
                        if r:
                            r = r[0]
                        else:
                            r = None
                if "value" in adapter:
                    r = draft1tool.resolve_eval(joborder, adapter["value"])
            if not r and "properties" in schema:
                r = {}
                for k, v in schema["properties"].items():
                    out = self.collect_outputs(v, joborder, outdir)
                    if out:
                        r[k] = out

        return r



import os
import random

class PathMapper(object):
    # Maps files to their absolute path
    def __init__(self, referenced_files, basedir):
        self._pathmap = {}
        for src in referenced_files:
            ab = src if os.path.isabs(src) else os.path.join(basedir, src)
            self._pathmap[src] = ab

    def mapper(self, src):
        return self._pathmap[src]


class DockerPathMapper(object):
    def __init__(self, referenced_files, basedir):
        self._pathmap = {}
        self.dirs = {}
        for src in referenced_files:
            abs = src if os.path.isabs(src) else os.path.join(basedir, src)
            dir, fn = os.path.split(abs)

            subdir = False
            for d in self.dirs:
                if dir.startswith(d):
                  subdir = True
                  break

            if not subdir:
                for d in list(self.dirs):
                    if d.startswith(dir):
                        # 'dir' is a parent of 'd'
                        del self.dirs[d]
                self.dirs[dir] = True

        prefix = "job" + str(random.randint(1, 1000000000)) + "_"

        names = set()
        for d in self.dirs:
            name = os.path.join("/tmp", prefix + os.path.basename(d))
            i = 1
            while name in names:
                i += 1
                name = os.path.join("/tmp", prefix + os.path.basename(d) + str(i))
            names.add(name)
            self.dirs[d] = name

        for src in referenced_files:
            abs = src if os.path.isabs(src) else os.path.join(basedir, src)
            for d in self.dirs:
                if abs.startswith(d):
                    self._pathmap[src] = os.path.join(self.dirs[d], abs[len(d)+1:])

    def mapper(self, src):
        return self._pathmap[src]



#!/usr/bin/env python

import os
import sys
import setuptools.command.egg_info as egg_info_cmd
import shutil

from setuptools import setup, find_packages

SETUP_DIR = os.path.dirname(__file__)
README = os.path.join(SETUP_DIR, 'README.rst')

try:
    import gittaggers
    tagger = gittaggers.EggInfoFromGit
except ImportError:
    tagger = egg_info_cmd.egg_info

# Remove the symlink and copy the schemas directory.
# This is a total hack, but older versions of setuptools
# won't follow symlinks or follow relative paths outside the
# source directory (ugh!)
restore = False
if os.path.islink("cwltool/schemas") and os.path.exists("../schemas"):
    os.unlink("cwltool/schemas")
    shutil.copytree("../schemas", "cwltool/schemas")
    restore = True

setup(name='cwltool',
      version='1.0',
      description='Common workflow language reference implementation',
      long_description=open(README).read(),
      author='Common workflow language working group',
      author_email='common-workflow-language@googlegroups.com',
      url="https://github.com/common-workflow-language/common-workflow-language",
      download_url="https://github.com/common-workflow-language/common-workflow-language",
      license='Apache 2.0',
      packages=["cwltool"],
      package_data={'cwltool': ['schemas/draft-1/*', 'schemas/draft-2/*']},
      install_requires=[
          'jsonschema >= 2.4.0',
          'requests',
          'PyYAML',
          'avro',
          'rdflib',
          'rdflib-jsonld'
        ],
      test_suite='tests',
      tests_require=[],
      entry_points={
          'console_scripts': [ "cwltool=cwltool.main:main" ]
      },
      zip_safe=False,
      cmdclass={'egg_info': tagger},
)

if restore:
    # Restore the symlink
    shutil.rmtree("cwltool/schemas")
    os.symlink("../../schemas", "cwltool/schemas")



#!/usr/bin/env python

import draft1tool
import draft2tool
import argparse
from ref_resolver import from_url
import jsonschema
import json
import os
import sys
import logging
import workflow
import validate

_logger = logging.getLogger("cwltool")
_logger.addHandler(logging.StreamHandler())


def printrdf(workflow, sr):
    from rdflib import Graph, plugin
    from rdflib.serializer import Serializer
    wf = from_url(workflow)
    g = Graph().parse(data=json.dumps(wf), format='json-ld', location=workflow)
    print(g.serialize(format=sr))

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("workflow", type=str)
    parser.add_argument("job_order", type=str, nargs="?", default=None)
    parser.add_argument("--conformance-test", action="store_true")
    parser.add_argument("--basedir", type=str)
    parser.add_argument("--outdir", type=str)
    parser.add_argument("--no-container", action="store_true", help="Do not execute jobs in a Docker container, even when specified by the CommandLineTool")
    parser.add_argument("--no-pull", default=False, action="store_true", help="Do not try to pull the Docker image")
    parser.add_argument("--dry-run", action="store_true", help="Do not execute")
    parser.add_argument("--verbose", action="store_true", help="Print more logging")
    parser.add_argument("--debug", action="store_true", help="Print even more logging")
    parser.add_argument("--print-rdf", action="store_true", help="Print corresponding RDF graph for workflow")
    parser.add_argument("--rdf-serializer", help="Output RDF serialization format (one of turtle (default), n3, nt, xml)", default="turtle")

    args = parser.parse_args()

    if args.verbose:
        logging.getLogger("cwltool").setLevel(logging.INFO)
    if args.debug:
        logging.getLogger("cwltool").setLevel(logging.DEBUG)

    if args.print_rdf:
        printrdf(args.workflow, args.rdf_serializer)
        return 0

    basedir = args.basedir if args.basedir else os.path.abspath(os.path.dirname(args.job_order))

    try:
        t = workflow.makeTool(from_url(args.workflow), basedir)
    except (jsonschema.exceptions.ValidationError, validate.ValidationException):
        _logger.exception("Tool definition failed validation")
        return 1

    try:
        job = t.job(from_url(args.job_order), basedir, use_container=(not args.no_container))
        if args.conformance_test:
            a = {"args": job.command_line}
            if job.stdin:
                a["stdin"] = job.stdin
            if job.stdout:
                a["stdout"] = job.stdout
            if job.generatefiles:
                a["generatefiles"] = job.generatefiles
            print json.dumps(a)
        else:
            (outdir, runjob) = job.run(dry_run=args.dry_run, pull_image=(not args.no_pull), outdir=args.outdir)
            _logger.info("Output directory is %s", outdir)
            print json.dumps(runjob)
    except (jsonschema.exceptions.ValidationError, validate.ValidationException):
        _logger.exception("Job order failed validation")
        return 1

    return 0

if __name__ == "__main__":
    sys.exit(main())



import subprocess
import os
import tempfile
import glob
import json
import yaml
import logging
import sys
import requests

_logger = logging.getLogger("cwltool")

class CommandLineJob(object):
    def run(self, dry_run=False, pull_image=True, outdir=None):
        if not outdir:
            if not dry_run:
                outdir = tempfile.mkdtemp()
            else:
                outdir = "/tmp"

        with open(os.path.join(outdir, "job.cwl.json"), "w") as fp:
            json.dump(self.joborder, fp)

        runtime = []
        env = {}

        if self.container and self.container.get("type") == "docker":
            found = False
            for ln in subprocess.check_output(["docker", "images", "--no-trunc", "--all"]).splitlines():
                try:
                    ln.index(self.container["imageId"])
                    found = True
                except ValueError:
                    pass

            if not found and pull_image:
                if "pull" in self.container:
                    cmd = ["docker", "pull", self.container["pull"]]
                    _logger.info(str(cmd))
                    if not dry_run:
                        subprocess.check_call(cmd, stdout=sys.stderr)
                        found = True
                elif "load" in self.container:
                    cmd = ["docker", "load"]
                    _logger.info(str(cmd))
                    if not dry_run:
                        if os.path.exists(self.container["load"]):
                            _logger.info("Loading docker image from %s", self.container["load"])
                            with open(self.container["load"], "rb") as f:
                                loadproc = subprocess.Popen(cmd, stdin=f, stdout=sys.stderr)
                        else:
                            _logger.info("Sending GET request to %s", self.container["load"])
                            req = requests.get(self.container["load"], stream=True)
                            n = 0
                            for chunk in req.iter_content(1024*1024):
                                n += len(chunk)
                                _logger.info(str(n))
                                loadproc.stdin.write(chunk)
                            loadproc.stdin.close()
                        rcode = loadproc.wait()
                        if rcode != 0:
                            raise Exception("Docker load returned non-zero exit status %i" % (rcode))
                        found = True

            if found:
                runtime = ["docker", "run", "-i"]
                for d in self.pathmapper.dirs:
                    runtime.append("--volume=%s:%s:ro" % (os.path.abspath(d), self.pathmapper.dirs[d]))
                runtime.append("--volume=%s:%s:rw" % (os.path.abspath(outdir), "/tmp/job_output"))
                runtime.append("--workdir=%s" % ("/tmp/job_output"))
                runtime.append("--user=%s" % (os.geteuid()))
                for t,v in self.environment.items():
                    runtime.append("--env=%s=%s" % (t, v))
                runtime.append(self.container["imageId"])
            else:
                raise Exception("Docker image %s not found" % (self.container["imageId"]))
        else:
            env = self.environment

        stdin = None
        stdout = None

        _logger.info("%s%s%s",
                     " ".join(runtime + self.command_line),
                     ' < %s' % (self.stdin) if self.stdin else '',
                     ' > %s' % (self.stdout) if self.stdout else '')

        if dry_run:
            return (outdir, {})

        if self.stdin:
            stdin = open(self.stdin, "rb")
        else:
            stdin = subprocess.PIPE

        os.chdir(outdir)

        if self.stdout:
            stdout = open(self.stdout, "wb")
        else:
            stdout = sys.stderr

        for t in self.generatefiles:
            with open(os.path.join(outdir, t), "w") as f:
                f.write(self.generatefiles[t])

        sp = subprocess.Popen(runtime + self.command_line,
                              shell=False,
                              close_fds=True,
                              stdin=stdin,
                              stdout=stdout,
                              env=env,
                              cwd=outdir)

        if stdin == subprocess.PIPE:
            sp.stdin.close()

        sp.wait()

        if stdin != subprocess.PIPE:
            stdin.close()

        if stdout != sys.stderr:
            stdout.close()

        return (outdir, self.collect_outputs(outdir))
